package jess;

import java.util.*;
import java.io.Serializable;

/** **********************************************************************
 * Agenda management.
 * <P>
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 */

class Agenda implements Serializable {
    
    private Object m_activationSemaphore = new String("ACTIVATION LOCK");
    private HeapPriorityQueue m_activations = new HeapPriorityQueue();
    private boolean m_halt = false;
    private int m_evalSalience = Rete.INSTALL;


    void setEvalSalience(int method) throws JessException {
        if (method < Rete.INSTALL || method > Rete.EVERY_TIME)
            throw new JessException("Agenda.setEvalSalience",
                                    "Invalid value", method);
        m_evalSalience = method;
    }


    int getEvalSalience() {
        return m_evalSalience;
    }
    
    void clear() {
        m_activations.clear();
        m_halt = false;
    }

    Object getActivationSemaphore() {
        return m_activationSemaphore;
    }
    
    Iterator listActivations() {
        return m_activations.iterator();
    }

    void addActivation(Activation a) throws JessException {

        if (m_evalSalience != Rete.INSTALL)
            a.getRule().evalSalience();

        m_activations.push(a);
        synchronized (m_activationSemaphore) {
            m_activationSemaphore.notify();
        }
    }

    void waitForActivations() {
        try {
            synchronized (m_activationSemaphore) {
                if (m_activations.isEmpty())
                    m_activationSemaphore.wait();
            }
        }
        catch (InterruptedException ie) { /* FALL THROUGH */ }
    }

    String setStrategy(Strategy s, Rete r) throws JessException {
        synchronized (m_activations) {
            Iterator e = m_activations.iterator();
            m_activations.clear();                
            String rv = m_activations.setStrategy(s).getName();
            while (e.hasNext()) {
                Activation a = (Activation) e.next(); 
                if (!a.isInactive())
                    r.addActivation(a);
            }
            return rv;
        }
    }

    void halt() {
        m_halt = true;
        synchronized (m_activationSemaphore) {
            m_activationSemaphore.notify();
        }
    }

    int runUntilHalt(Rete r) throws JessException {
        int count = 0;
        do {
            count += run(r);
            if (m_halt)
                break;
            waitForActivations();
        } while (!m_halt);

        return count;
    }

    synchronized int run(Rete r) throws JessException {
        int i=0, j;
        do {
            j = run(Integer.MAX_VALUE, r);
            i += j;
        } while (j > 0 && !m_halt);
        return i;
    }

    synchronized int run(int max, Rete r) throws JessException {
        int n = 0;
        int size = 0;
        m_halt = false;
    
        while (!m_activations.isEmpty() && !m_halt && n < max) { 
            Thread.yield();
        
            if (!m_activations.isEmpty()) {
                Activation a = m_activations.pop();
                if (a == null)
                    return n;
                            
                if (!a.isInactive()) {
                    a.setSequenceNumber(++n);
                    r.broadcastEvent(JessEvent.DEFRULE_FIRED, a);
                    try {
                        r.aboutToFire(a);
                        a.fire();
                    } finally {
                        r.justFired(a);
                    }
                }
            }

            if (m_evalSalience == Rete.EVERY_TIME) {
                setStrategy(m_activations.getStrategy(), r);
            }
        }
        return n;
    }
}
