package jess;

import java.io.Serializable;
import java.util.*;
import java.lang.reflect.*;
import java.beans.*;

/** **********************************************************************
 * The current list of definstances.
 * <P>
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 * @author Ernest J. Friedman-Hill
 */

class DefinstanceList implements Serializable, PropertyChangeListener {

    // Keys are Jess defclass names, elements are Java class names
    private HashMap m_javaClasses = new HashMap(101);
    // Keys are objects to match, elements are the facts that represent them.
    private HashMap m_definstances = new HashMap(101);
    // Keys are objects to match, elements are the Jess class names
    private HashMap m_jessClasses = new HashMap(101);

    private Rete m_engine;

    DefinstanceList(Rete engine) {
        m_engine = engine;
    }

    void clear() {
        m_javaClasses.clear();
        m_definstances.clear();
        m_jessClasses.clear();      
    }

    void reset() throws JessException {
        Iterator e = m_definstances.keySet().iterator();
        while (e.hasNext())
            createFact(e.next(), null, null, true);
    }
    
    Iterator listDefinstances() {
        return m_definstances.keySet().iterator();
    }

    String jessNameToJavaName(String s) {
        return (String) m_javaClasses.get(s);
    }

    void mapDefclassName(String javaName, String className) {
        m_javaClasses.put(javaName, className);
    }

    Fact createFact(Object o, String changedName, Object newValue)
        throws JessException {
        return createFact(o, changedName, newValue, false);
    }
    
    synchronized Fact createFact(Object o, String changedName,
                                  Object newValue,
                                  boolean reset) 
        throws JessException {
        // Synthesize a fact for this object; remember it for
        // later retraction
        Fact fact = (Fact) m_definstances.get(o);
        boolean doGetValue = (newValue == null || changedName == null);
        boolean keepID = false;
        try {
            
            if (fact == null) {
                fact = new Fact((String) m_jessClasses.get(o), m_engine);
                fact.setSlotValue("OBJECT", new Value(o));
                fact.setShadow(true);
                m_definstances.put(o,fact);
                
            } else {
                if (!reset)
                    keepID = true;
                m_engine.retract(fact);
            }
            
            
            Deftemplate deft = fact.getDeftemplate();
            Object [] args = new Object[] {};
            boolean changedSomething = false;
            for (int i=0; i<deft.getNSlots(); i++) {
                if (deft.getSlotName(i).equals("OBJECT"))
                    continue;
                SerializablePropertyDescriptor pd =
                    (SerializablePropertyDescriptor)
                    deft.getSlotDefault(i).
                    externalAddressValue(m_engine.getGlobalContext());
                String name = pd.getName();
                
                // changedName is null if multiple props changed or
                // completely new fact is desired
                if (changedName != null && !name.equals(changedName))
                    continue;
                
                Method m = pd.getReadMethod();
                Class rt = m.getReturnType();
                
                if (doGetValue)
                    newValue = m.invoke(o, args);
                
                Value oldV = fact.getSlotValue(name);
                
                Value newV = ReflectFunctions.objectToValue(rt, newValue);
                if (!oldV.equals(newV)) {
                    fact.setSlotValue(name, newV);
                }
                
                changedSomething = true;            
            }      
            if (changedName != null && !changedSomething)
                throw new JessException("DefinstanceList.createFact",
                                        "Property not found:",
                                        changedName);
            
        } catch (InvocationTargetException ite) {
            throw new JessException("DefinstanceList.createFact",
                                    "Called method threw an exception",
                                    ite.getTargetException()); 
            
        } catch (IllegalAccessException iae) {
            throw new JessException("DefinstanceList.createFact",
                                    "Method is not accessible",
                                    iae);
        } catch (IllegalArgumentException iae) {
            throw new JessException("DefinstanceList.createFact",
                                    "Invalid argument", iae);
        } finally {
            // Regardless of what happens, we need to assert this
            // now, since retracted it before!
            if (keepID)
                fact = m_engine.assertKeepID(fact);        
            else
                fact = m_engine.assert(fact);        
        }
        return fact;
    }

    Value definstance(String jessTypename, Object object, boolean dynamic)
        throws JessException {
        try {
            String javaTypename = jessNameToJavaName(jessTypename);

            if (javaTypename == null)
                throw new JessException("DefinstanceList.definstance",
                                        "Unknown object class",
                                        jessTypename);
                        
            // Make sure we're not already matching on this object
            if (m_definstances.get(object) != null)
                return new FactIDValue(null);
            
            
            if (!Class.forName(javaTypename).
                isAssignableFrom(object.getClass()))
                throw new JessException("DefinstanceList.definstance",
                                        "Object is not instance of",
                                        javaTypename);
            
            if (dynamic) {
                // Add ourselves to the object as a PropertyChangeListener 
                
                Class pcl = Class.forName("java.beans.PropertyChangeListener");
                Method apcl =
                    object.getClass().getMethod("addPropertyChangeListener",
                                           new Class[] { pcl });    
                apcl.invoke( object, new Object[] { this });
            }
            
            m_jessClasses.put(object, jessTypename);
            Fact fact = createFact(object, null, null);
            
            return new FactIDValue(fact);

        } catch (InvocationTargetException ite) {
            throw new JessException("DefinstanceList.definstance",
                                    "Cannot add PropertyChangeListener",
                                    ite.getTargetException());

        } catch (NoSuchMethodException nsm) {
            throw new JessException("DefinstanceList.definstance",
                                    "Obj doesn't accept " + 
                                    "PropertyChangeListeners",
                                    nsm);
        } catch (ClassNotFoundException cnfe) {
            throw new JessException("DefinstanceList.definstance",
                                    "Class not found", cnfe);
        } catch (IllegalAccessException iae) {
            throw new JessException("DefinstanceList.definstance",
                                    "Class or method is not accessible",
                                    iae);
        }
    }

    Value undefinstance(Object o) throws JessException {
        Fact f = (Fact) m_definstances.get(o);
        if (f != null)
            m_engine.retract(f);
        m_definstances.remove(o);
        
        try {
            Method apcl =
                o.getClass().getMethod("removePropertyChangeListener",
                                       new Class[] { 
                                           PropertyChangeListener.class
                                       });    
            apcl.invoke( o, new Object[] { this });

        } catch (Exception e) { /* whatever */ }
        
        if (m_jessClasses.remove(o) == null)
            return Funcall.FALSE;
        else
            return Funcall.TRUE;

    }        
    
    public synchronized void propertyChange(PropertyChangeEvent pce) {
        Object o = pce.getSource();
        
        try {
            String s = (String) m_jessClasses.get(o);
            if (s != null)
                createFact(o, pce.getPropertyName(), pce.getNewValue());
            
        } catch (JessException re) {
            System.out.println("Async Error: " + re);
            if (re.getNextException() != null)
                re.getNextException().printStackTrace();
        }
    }

    Value defclass(String jessName, String clazz, String parent)
        throws JessException {
    
        try {
            Deftemplate dt;
            if (parent != null) {
                dt = m_engine.findDeftemplate(parent);
                if (dt == null) 
                    throw new JessException("defclass",
                                            "No such parent template: ",
                                            parent);
                dt = new Deftemplate(jessName, "$JAVA-OBJECT$ " + clazz, dt);
            }
            else
                dt = new Deftemplate(jessName, "$JAVA-OBJECT$ " + clazz);
            
            Class c = m_engine.findClass(clazz);
            mapDefclassName(jessName, c.getName());
            
            // Make all the readable 'bean properties' into slots
            PropertyDescriptor [] props =
                ReflectFunctions.getPropertyDescriptors(c);
            
            // Sort them first
            for (int i=0; i<props.length-1; i++)
                for (int j=i+1; j<props.length; j++)
                    if (props[i].getName().compareTo(props[j].getName()) > 0) {
                        PropertyDescriptor temp = props[i];
                        props[i] = props[j];
                        props[j] = temp;
                    }
            
            
            // TODO: should set proper slot types
            for (int i=0; i<props.length; i++) {
                Method m = props[i].getReadMethod();
                if (m == null)
                    continue;
                String name = props[i].getName();
                Class rt = m.getReturnType();
                Value slot =
                    new Value(new SerializablePropertyDescriptor(c,
                                                                 props[i]));
                if (rt.isArray())
                    dt.addMultiSlot(name, slot);
                else
                    dt.addSlot(name, slot, "ANY");

            }
            
            // Last slot is special - it holds the active instance
            dt.addSlot("OBJECT", Funcall.NIL, "OBJECT");
            
            // Install our synthetic deftemplate
            m_engine.addDeftemplate(dt);
            
            return new Value(c.getName(), RU.ATOM);
        } catch (ClassNotFoundException cnfe) {
            throw new JessException("defclass", "Class not found:", cnfe);
        } catch (IntrospectionException ie) {
            throw new JessException("defclass", "Introspection error:", ie);
        }
    }

}

