package jess;

import java.io.Serializable;
import java.util.*;
import java.io.Writer;
import java.io.IOException;

/** **********************************************************************
 * The current list of facts.
 * <P>
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 * @author Ernest J. Friedman-Hill
 */


class FactList implements Serializable {

    private HashMap m_table = new HashMap();

    private ArrayList m_factsToAssert = new ArrayList(),
        m_factsToRetract = new ArrayList();
    private int m_time = 0;
 
    private static Value m_rewrite;
    static {
        try {
            m_rewrite = new Value("modify-n", RU.ATOM);
        }
        catch (JessException re) { /* can't happen */ }
    }
   
    void clear() {
        m_table.clear();
        m_time = 0;
        m_factsToAssert.clear();
        m_factsToRetract.clear();
        m_nextFactId = 0;
    }

    int getTime() {
        return m_time;
    }

    void assignTime(Fact fact) {
        fact.updateTime(++m_time);
    }

    int doPreAssertionProcessing(Fact f) throws JessException {
        Object o = m_table.get(f);
        return (o == null) ? 0 : -1;
    }
  
    Fact findFactByID(int id) throws JessException {
        synchronized (m_table) {
            for (Iterator it = m_table.keySet().iterator(); it.hasNext();) {
                Fact f = (Fact) it.next();
                if (f.getFactId() == id)
                    return f;
            }
            return null;
        }
    }

    public void ppFacts(String name, Writer output) throws IOException {
        for (Iterator e = listFacts(); e.hasNext();) {
            Fact f = (Fact) e.next();
            if (name == null || f.getName().equals(name)) {
                output.write(f.toString());
                output.write("\n");
            }
        }
    }
    
    public void ppFacts(Writer output) throws IOException {
        ppFacts(null, output);
    }

    public Iterator listFacts() {        
        return new SortedIterator(m_table);
    }

    /**
     * Successively incremented ID for asserted facts.
     */
    private Object m_idLock = "LOCK";
    private int m_nextFactId;

    private int nextFactId() {
        synchronized (m_idLock) {
            return m_nextFactId++;
        }
    }

    /**
     * Facts generated by LHS actions, not asserted until RHS time.
     */
  

    void setPendingFact(Fact fact, boolean assert) {
        if (assert)
            m_factsToAssert.add(fact);
        
        else {
            m_factsToRetract.add(fact);
        }
    }

    // Process any facts that were asserted by rule LHS processing -- i.e.,
    // for backwards chaining.
    void processPendingFacts(Rete r) throws JessException {
        synchronized (m_factsToAssert) {
            while (m_factsToAssert.size() > 0) {
                Fact f = (Fact) m_factsToAssert.get(0);
                m_factsToAssert.remove(0);
                _assert(f, true, r);
            }
        }
        
        synchronized (m_factsToRetract) {
            while (m_factsToRetract.size() > 0) {
                _retract((Fact) m_factsToRetract.get(0), r);
                m_factsToRetract.remove(0);
            }
        }
    }
    
    public Fact assert(Fact f, Rete r) throws JessException {
        Fact fact = _assert(f, true, r);
        if (fact != null)
            processPendingFacts(r);
        return fact;
    }

    Fact assertKeepID(Fact f, Rete r) throws JessException {
        Fact fact = _assert(f, false, r);
        if (fact != null)
            processPendingFacts(r);
        return fact;
    }

    private Fact _assert(Fact f, boolean newID, Rete r) throws JessException {
        synchronized (m_table) {
            // find any old copy
            if (r.doPreAssertionProcessing(f) != 0)
                return null;
        
            // insert the new fact
            if (newID)
                f.setFactId(nextFactId());

            r.broadcastEvent(JessEvent.FACT, f);        
                
            m_table.put(f, f);
        
            // Send it to the Rete network
            r.processToken(RU.ADD, f);        
            return f;
        }
    }

    /**
     * Retract a fact. 
     * @param f A Fact object. Doesn't need to be the actual object
     * that appears on the fact-list; can just be a Fact that could
     * compare equal to one.
     * @exception JessException If anything goes wrong.
     */

    Fact retract(Fact f, Rete r) throws JessException {
        synchronized (m_table) {
            Fact ff;
            if ((ff = (Fact) m_table.get(f)) != null) {
                _retract(ff, r);
                processPendingFacts(r);
                return ff;
            }
            else
                return null;
        }
    }

    // f should be the actual fact
    private Fact _retract(Fact f, Rete r) throws JessException  {
        synchronized (m_table) {
            r.broadcastEvent(JessEvent.FACT | JessEvent.REMOVED, f);
            m_table.remove(f);
            r.processToken(RU.REMOVE, f);            
            return f;
        }
    }

    Fact _modify(ValueVector vv, Context context, Rete r) throws JessException {
        Fact fact = (Fact) vv.get(1).externalAddressValue(context);
        
        if (fact.getFactId() == -1)
            throw new JessException("modify", "no such fact", -1);

        Deftemplate dt = fact.getDeftemplate();

        //////////////////////////////////////////////////
        // First find out of this is a LHS object, not an ordinary fact
        // If so, call the appropriate mutators...
        //////////////////////////////////////////////////

        if (fact.isShadow()) {
            Value ov = fact.getSlotValue("OBJECT");
            Funcall fc = new Funcall("set", r).
                arg(ov).arg(new Value("set", RU.STRING));
            fc.setLength(4);
            
            for (int i= 2; i < vv.size(); i++) {
                
                // fetch the slot, value subexp, stored as a List
                ValueVector svp = vv.get(i).listValue(context);
                String slotName = getSlotName(svp, context);
                
                fc.set(new Value(slotName, RU.ATOM), 2);
                
                int type = dt.getSlotType(dt.getSlotIndex(slotName));
                fc.set(getSlotValue(svp, context, type), 3);
                
                fc.execute(r.getGlobalContext());
            }
            return fact;
        }

        //////////////////////////////////////////////////
        // Not an object - just a fact.
        //////////////////////////////////////////////////

        // Remove the old fact from the Rete network
    
        r.processToken(RU.REMOVE, fact);            

        if (vv.get(0).equals(m_rewrite)) {
            // We know the slot names now hold slot numbers.

            for (int i= 2; i < vv.size(); i++) {
                ValueVector svp = vv.get(i).listValue(context);
                int[] ia = (int[]) svp.get(0).externalAddressValue(context);
                fact.set(getSlotValue(svp, context, ia[1]), ia[0]);
            }

        } else {
            // We've never done this one before. Besides modifying
            // the fact, we rewrite the function call by storing slot offsets
            // in place of slot names.
        
            for (int i= 2; i < vv.size(); i++) {
                
                // fetch the slot, value subexp, stored as a List 
                ValueVector svp = vv.get(i).listValue(context);
                String slotName = getSlotName(svp, context);
                int idx = dt.getSlotIndex(slotName);
                int type = dt.getSlotType(idx);
                
                // Set the value in the fact
                fact.setSlotValue(slotName, getSlotValue(svp, context, type));
                
                // modify the funcall to contain the index and type
                svp.set(new Value((Object) new int[] {idx, type}), 0);
            }

            // Mark this funcall as rewritten
            vv.set(m_rewrite, 0);
        }
        // Add the modified fact back into the Rete network
        r.broadcastEvent(JessEvent.FACT | JessEvent.MODIFIED, fact);        
        
        if (doPreAssertionProcessing(fact) == 0) {
            r.processToken(RU.ADD, fact);            
            return fact;

        } else {
            m_table.remove(fact);

            r.processToken(RU.REMOVE, fact);            
            return Fact.getNullFact();
        }    
    }

    private String getSlotName(ValueVector svp, Context c)
        throws JessException {
        
        return svp.get(0).stringValue(c);
    }
 
    private Value getSlotValue(ValueVector svp, Context c, int type)
        throws JessException {

        if (type == RU.SLOT) {
            Value v = svp.get(1).resolveValue(c);
            while (v.type() == RU.LIST)
                v = v.listValue(c).get(0).resolveValue(c);
        
            return v;

        } else { // MULTISLOT 
            ValueVector vv = new ValueVector();
            for (int i=1; i<svp.size(); i++) {
                Value listItem = svp.get(i).resolveValue(c);
                if (listItem.type() == RU.LIST) {
                    ValueVector sublist = listItem.listValue(c);
                    for (int j=0; j<sublist.size(); j++)
                        vv.add(sublist.get(j).resolveValue(c));
                }   
                else
                    vv.add(listItem);
            }
            return new Value(vv, RU.LIST);
        }
    }
}

class FactComparator implements Comparator, Serializable {
    public int compare(Object o1, Object o2) {
    return ((Fact) o1).getFactId() - ((Fact) o2).getFactId();
}
}

class SortedIterator implements Iterator, Serializable {
    private Object[] m_arr;
    private int m_index;
    
    SortedIterator(Map map) {
        m_arr = map.keySet().toArray();
        Arrays.sort(m_arr, new FactComparator());
    }
    public boolean hasNext() {
        return m_index < m_arr.length;
    }
    public Object next() {
        return m_arr[m_index++];
    }
    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }
}
