package jess;

import java.io.Serializable;
import java.util.*;

/** 
 * A group on the LHS of a rule, like an AND, OR, NOT, EXPLICIT, or
 * other special CE. 
 * (C) 2001 E.J. Friedman-Hill and Sandia National Laboratories
 */

class Group implements Groupable, Serializable {
    private String m_name;
    private boolean m_unique, m_explicit;
    private int m_negcnt=0;
    private GroupableVector m_data;
    private boolean m_unary = false;

    static final String
        AND="and", UNIQUE="unique", EXPLICIT="explicit",
        NOT="not", EXISTS="exists", TEST="test", OR="or";

    Group(String name) {
        m_data = new GroupableVector();
        m_name = name;

        if (m_name.equals(UNIQUE)) {
            m_unique = true;
            m_unary = true;
        }

        else if (m_name.equals(EXPLICIT)) {
            m_explicit = true;
            m_unary = true;
        }

        else if (m_name.equals(NOT)) {
            m_negcnt = 1;
            m_unary = true;
        }

        else if (m_name.equals(EXISTS)) {
            m_negcnt = 2;
            m_unary = true;
        }    
    }

    public String getName() {
        return m_name;
    }

    void add(Groupable g) throws JessException {
        if (m_data.size() > 0 && m_unary)
            throw new JessException("Group.add",
                                    "CE is a unary modifier", m_name);

        // Flatten nested ORs immediately.
        if (m_name.equals(OR) && g.getName().equals(OR)) {
            Group grp = (Group) g;
            for (int i=0; i<grp.m_data.size(); i++)
                m_data.add(grp.m_data.get(i));
        }
        // Otherwise keep the tree-like structure
        else
            m_data.add(g);
    
        if (m_explicit)
            g.setExplicit(true);
    
        if (m_unique)
            g.setUnique();    
    }
  
    int updateNegcnt(int negcnt) {
        m_negcnt += negcnt;
        return m_negcnt;
    }

    int getNegcnt() {
        return m_negcnt;
    }

    public void setExplicit(boolean exp) {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setExplicit(exp);
    }

    public void setUnique() {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setUnique();
    }

    public void setBoundName(String name) throws JessException {
        if (m_name.equals(NOT) ||
            m_name.equals(TEST) ||
            (m_data.size() > 1 && !m_name.equals(OR)))
            throw new JessException("Group.setBoundName",
                                    "This CE can't be bound to a variable",
                                    m_name);

        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setBoundName(name);
    }

    static boolean isGroupName(String s) {
        return
            s.equals(NOT) ||
            s.equals(AND) ||
            s.equals(OR) ||
            s.equals(EXISTS) ||
            s.equals(UNIQUE) ||
            s.equals(EXPLICIT);     
    }

    public String toString() {
        StringBuffer sb = new StringBuffer("(");
        sb.append(m_name);
        for (int i=0; i<m_data.size(); i++) {
            sb.append('\n');
            sb.append(' ');
            sb.append(m_data.get(i));
        }
        sb.append(")");
        return sb.toString();
    }

    private int countNumberOfBranches(Groupable[] lhss) throws JessException {

        for (int i=0; i<lhss.length; i++)
            lhss[i] = m_data.get(i).simplify();

        int product;
        if (m_name.equals(OR)) {
            product = 0;
            for (int i=0; i<lhss.length; i++) {
                product += lhss[i].getGroupSize();
            }
        }
        else {
            product = 1;
            for (int i=0; i<lhss.length; i++) {
                product *= lhss[i].getGroupSize();
            }
        }
        return product;
    }


    public Groupable simplify() throws JessException {

        Groupable[] lhss = new Groupable[m_data.size()];
        int product = countNumberOfBranches(lhss);

        if (product == 1) {
            Group g = new Group(OR);
            g.add(this);
            return g;
        }

        // Each element in rv is a new LHS being built.
        Group g = new Group(OR);

        for (int i=0; i<product; i++)
            g.add(new Group(AND));
   
        if (m_name.equals(OR)) {        
            // If we're an OR, then we're just the OR of all the children.
            // Each child in lhss is just a list of one or more OR-options
            int index=0;
            for (int i=0; i<lhss.length; i++)
                for (int j=0; j<lhss[i].getGroupSize(); j++)
                    ((Group) g.getGroupable(index++))
                        .add(lhss[i].getGroupable(j));          
        }
        else {
            // Otherwise, it's a little more complex. We need to
            // distribute our OR children. "Repeat" is the number of
            // consecutive LHSs in the list a given sub-OR element
            // will be added to. It will get smaller and smaller and
            // end up as 1 for the last OR.

            int copies = product;
            for (int i=0; i<lhss.length; i++) {
                // If there's only one option in a branch, add it to every LHS.
                if (lhss[i].getGroupSize() == 1) {
                    for (int j=0; j<product; j++)
                        ((Group) g.getGroupable(j)).add(lhss[i].getGroupable(0));
                }
            
                // If there are multiple options, we need to spread them
                // out.  We'll make "copies" copies of each option,
                // repeating the overall pattern "repeat" times. Each time
                // we encounter another OR in lhss, we divide "copies" by
                // the number of branches. If lhss looks like
                // 
                //  A C D 
                //  B   E
                //
                // Then the first and third entries are ORs, and we want rv to look like
                //
                //  A A B B  copies == 2, repeat == 1
                //  C C C C
                //  D E D E  copies == 1, repeat == 2
            
                else {
                    copies /= lhss[i].getGroupSize();  
                    int repeat = product/(copies * lhss[i].getGroupSize());
                    int index=0;
                    for (int j=0; j<repeat; j++)
                        for (int k=0; k<lhss[i].getGroupSize(); k++)
                            for (int m=0; m<copies; m++)
                                ((Group) g.getGroupable(index++)).
                                    add(lhss[i].getGroupable(k));
                }
            }
        }

        return g;
    
    }

    public void addToLHS(HasLHS l) throws JessException {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).addToLHS(l);
    }

    public int getGroupSize() {
        return m_data.size();
    }
    
    public Groupable getGroupable(int i) {
        return m_data.get(i);
    }  

    private static class GroupableVector {
        private Groupable[] m_data = new Groupable[1];
        private int m_nData;
    
        void add(Groupable g) {
      
            if (m_data.length == m_nData) {
                Groupable[] temp = new Groupable[m_nData * 2];
                System.arraycopy(m_data, 0, temp, 0, m_nData);;
                m_data = temp;      
            }
            m_data[m_nData++] = g;
        }
    
        Groupable get(int i) {
            return m_data[i];
        }
        int size() {
            return m_nData;
        }
    }
}








